#' @importFrom fields rdist
#' @importFrom parallel mclapply
#' @importFrom Rcpp evalCpp
#' @importFrom truncnorm rtruncnorm
#' @import lattice


#########################################################################
## independent simulation
##
## inputs:
##
## dim := number of dimensions
## n   := number of observations
## p   := number of voxels
## m   := number of knots on each axis
## s   := location of voxels
## r   := radius of true beta
## 
## outputs:
## 
## Y     := n-vector of responses
## X     := n x p matrix of covariates
## beta  := simulated true beta
## B     := p x m matrix of kernel basis functions
## ADJ   := m x m adjacency matrix for the kernel knots
## voxel := voxel of the vector space 
#########################################################################
## Define the spatial adjacency matrix

## 1D

ADJ1D <- function(m){
  A <- 0*diag(m)
  for(j in 2:m){
    A[j,j-1]<-A[j-1,j]<-1
  }
  A}

# 2D

find_neighbors_2D = function(grids,radius = 0.05){
  names(grids) = c("x","y")
  x=unique(grids$x)
  y=unique(grids$y)
  x_neighbors = mclapply(x,function(a) which(abs(grids$x-a)<radius),mc.cores = 7)
  y_neighbors = mclapply(y,function(a) which(abs(grids$y-a)<radius),mc.cores = 7)
  
  names(x_neighbors) = paste(x)
  names(y_neighbors) = paste(y)
  
  neighbors=mclapply(1:nrow(grids),function(i){
    idx = intersect(x_neighbors[[paste(grids$x[i])]],y_neighbors[[paste(grids$y[i])]])
    #distance = sqrt((grids$x[idx]-grids$x[i])^2+(grids$y[idx]-grids$y[i])^2)
    return(idx)
    #temp[which(distance<radius)]
  },mc.cores = 7)
  return(neighbors)}

ADJ2D <- function(mm, AA){
  A <- 0 * diag(mm)
  for (j in 1:mm){
    A[j, AA[[j]]] <-1		
  }
  diag(A) <- 0
  A
}

## 3D

find_neighbors_3D = function(grids,radius = 3){
  names(grids) = c("x","y","z")
  x=unique(grids$x)
  y=unique(grids$y)
  z=unique(grids$z)
  x_neighbors = mclapply(x,function(a) which(abs(grids$x-a)<radius),mc.cores = 7)
  y_neighbors = mclapply(y,function(a) which(abs(grids$y-a)<radius),mc.cores = 7)
  z_neighbors = mclapply(z,function(a) which(abs(grids$z-a)<radius),mc.cores = 7)
  
  names(x_neighbors) = paste(x)
  names(y_neighbors) = paste(y)
  names(z_neighbors) = paste(z)
  
  neighbors=mclapply(1:nrow(grids),function(i){
    idx = intersect(x_neighbors[[paste(grids$x[i])]],y_neighbors[[paste(grids$y[i])]])
    idx = intersect(idx,z_neighbors[[paste(grids$z[i])]])
    #distance = sqrt((grids$x[idx]-grids$x[i])^2+(grids$y[idx]-grids$y[i])^2+(grids$z[idx]-grids$z[i])^2)
    return(idx)
    #temp[which(distance<radius)]
  },mc.cores = 7)
  return(neighbors)}

ADJ3D <- function(mm, AA){
  A <- 0 * diag(mm)
  for (j in 1:mm){
    A[j, AA[[j]]] <-1		
  }
  diag(A) <- 0
  A
}
#########################################################################
sim.STGP = function(n,p,m,dim=1,r=c(0.1,0.5)){
  ## 1D
  
  if (dim==1){
    
    ## settings
    
    s <- 1:p /(p+1)
    S <- s
    
    ## True beta
    
    beta <- ifelse(S > r[1] & S <= r[2],1,0)
    
    ## Generate data
    
    X    <- matrix(rnorm(n*p),n,p)
    Y    <- rnorm(n,X%*%beta,1)
    
    ## Set up the basis functions:
    
    knots <- seq(0,1,length=m)
    bw    <- min(dist(knots))
    
    A     <- ADJ1D(m)
    
    D     <- rdist(s,knots)
    B     <- exp(-0.5*(D/bw)^2)
    B     <- ifelse(D>2.5*bw,0,B)
  }
  
  ## 2D
  
  if (dim==2){
    
    ## Settings
    
    p2 <- p*p
    s  <- (1:p) / (p+1)            
    S  <- expand.grid(x=s, y=s)
    m2 <- m*m
    
    ## True beta
    
    beta = ifelse((S$x-0.5)^2+(S$y-0.5)^2< r^2,1,0)
    
    ## Generate data
    
    X    <- matrix(rnorm(n*p2),n,p2)
    Y    <- rnorm(n,X%*%beta,1)
    
    ## Set up the basis functions:
    
    knots <- seq(0,1,length=m)
    Grid  <- expand.grid(x = knots, y = knots)
    bw    <- min(dist(Grid))
    
    AA    <- find_neighbors_2D(Grid, bw+0.05)
    A     <- ADJ2D(m2, AA)
    
    D     <- rdist(S,Grid)
    B     <- exp(-0.5*(D/bw)^2)
    B     <- ifelse(D>2.5*bw,0,B)
  }
  
  ## 3D
  
  if (dim==3){
    
    ## Settings
    
    p3 <- p^3
    s  <- (1:p) / (p+1)            
    S  <- expand.grid(x=s, y=s, z=s)
    m3 <- m^3
    
    ## True beta
    
    beta = ifelse((S$x-0.5)^2+(S$y-0.5)^2+(S$z-0.5)^2< r^2,1,0)
    
    ## Generate data
    
    X    <- matrix(rnorm(n*p3),n,p3)
    Y    <- rnorm(n,X%*%beta,1)
    
    ## Set up the basis functions:
    
    knots <- seq(0,1,length=m)
    Grid  <- expand.grid(x = knots, y = knots, z = knots)
    bw    <- min(dist(Grid))
    
    AA    <- find_neighbors_3D(Grid, bw+0.05)
    A     <- ADJ3D(m3, AA)
    
    D     <- rdist(S,Grid)
    B     <- exp(-0.5*(D/bw)^2)
    B     <- ifelse(D>2.5*bw,0,B)
  }
  
  output <- list(Y=Y,
                 X=X,
                 beta=beta,
                 B=B,
                 ADJ=A,
                 voxel=S)
  
  return (output)
}

#######################################################################




#########################################################################
## STGP function
##
## inputs
##
## Y      := n-vector of responses
## X      := n x p matrix of covariates
## B      := p x L matrix of kernel basis functions
## ADJ    := L x L adjacency matrix for the kernel knots
## t_min  := prior for the threshold is Unif(t_min,t_max)
## t_max  :=
## binary := is the response binary?
## Xp     := n_pred x p matrix of covariates for test set prediction
## ae,be  := If Y is Gaussian its variance has InvG(ae,be) prior
## ar,br  := CAR dependence parameter has a beta(ar,br) prior
## eps    := beta[j]'s prior variance is 1/eps[j]
## iters  := Number of MCMC iterations
## burn   := number if discard as burn-in
## thin   := degree of thinning
## update := Plot current value every update seconds
##
##
## Gaussian model
##
## Y[i]  ~ N(beta[1] + beta[2]*sum(X[i,]*b),sigma^2)
## Z    <- B%*%alpha
## b    <- g(Z,thresh,w) where w are weights to ensure equal variance
##                        and g is the soft-thresholding function(below)  
## alpha ~ CAR with adjacency matrix ADJ
##
##
## Outputs   
##
## beta.mn  := Posterio mean of beta[2]*b
## beta.var := Posterior variance of beta[2]*b
## beta_nz  := Posterior probability beta[2]*b is non-zero
## fit.mn   := Posterior mean of test set Y
## fit.var  := Posterior variance of test set Y
## keepers  := Posterior draws for other parameters
##
###############################################################################
## Scale the weights to sum to one at each site

get.scale <- function(B,ADJ,rho){
  Q <- diag(colSums(ADJ))-rho*ADJ
  w <- rowSums((B%*%GaussInverse(Q))*B)
  return(1/sqrt(w))}

#log likelihood function:

log.like<-function(Y,beta,Xb,sige){
  mn <- beta[1]+beta[2]*Xb
  ll <- sum(dnorm(Y,mn,sige,log=TRUE))
  return(ll)}

## Update Threshold

updateThresh <- function(t_min,t_max,X,Y,Z,w,beta,sige,taua,thresh,b,Xb){
  cant    <- runif(1,t_min,t_max)
  canb    <- taua * g_thresh(Z,cant,w)
  canXb   <- X %*% canb
  diffLL  <- diffLoglike(Y,canXb,Xb,beta,sige)
  if(log(runif(1)) < diffLL){
    thresh  <- cant
    b       <- canb
    Xb      <- canXb
  }
  R <- list(thresh=thresh,b=b,Xb=Xb)
  return(R)
}

## update taua

updateTaua <- function(X,Y,Z,thresh,w,beta,sige,taua,b,Xb){
  can.taua <- rtruncnorm(1,a=1,b=Inf,mean=0,sd=1)
  can.b    <- b * (can.taua / taua)
  can.Xb   <- Xb * (can.taua / taua)
  diffLL   <- diffLoglike(Y,can.Xb,Xb,beta,sige)
  if(log(runif(1)) < diffLL){
    taua <- can.taua
    b    <- can.b
    Xb   <- can.Xb
  }
  R1 <- list(taua=taua,b=b,Xb=Xb)
  return(R1)
}

# update rho

updateRho <- function(X,Y,Z,alpha,adj1,adj2,nr,weight_grid,thresh,beta,sige,taua,rho.grid,eee,
                      rho.g,rho,w,b,Xb){
  SS2       <- 2*sum(alpha[adj1]*alpha[adj2])
  can.rho.g <- sample(1:nr,1)
  can.w     <- weight_grid[,can.rho.g]
  can.b     <- taua * g_thresh(Z,thresh,can.w)
  can.Xb    <- X %*% can.b
  diffLL    <- diffLoglike(Y,can.Xb,Xb,beta,sige)
  diff      <- diffLL +
    0.5*taua*rho.grid[can.rho.g]*SS2+eee[can.rho.g]-
    0.5*taua*rho.grid[rho.g]*SS2-eee[rho.g]
  if(log(runif(1)) < diff){
    rho.g <- can.rho.g
    rho   <- rho.grid[rho.g]
    w     <- weight_grid[,rho.g]
    b     <- can.b
    Xb    <- can.Xb
  }
  R2 <- list(rho.g=rho.g,rho=rho,w=w,b=b,Xb=Xb)
}

###############################################################################
STGP<-function(Y,X,B,ADJ,t_min,t_max,binary=FALSE,Xp=NULL,
               ae=0.1,be=0.1,ar=10,br=1,eps=c(0.1,10),
               iters=5000,burn=1000,thin=2,show_iters=5){
  
  
  #Bookkeeping
  
  L <- ncol(B)
  n <- length(Y)
  p <- ncol(X)
  
  if(binary){
    LOW  <- ifelse(Y==1,0,-Inf)
    HIGH <- ifelse(Y==1,Inf,0)
    Y    <- ifelse(Y==1,2,-2)
  }
  
  #initial values 
  
  ols   <- rep(0,p)
  for(j in 1:p){
    ols[j]<-fastLM(Y, X[,j])$beta
  }  
  ols   <- 5*ols/max(ols)
  alpha <- GaussInverse(t(B)%*%B+.01*diag(L))%*%t(B)%*%ols
  
  beta   <- c(0,1)
  taue   <- 1
  taua   <- 1
  thresh <- (t_min+t_max)/2
  
  # Keep track of stuff
  
  B1 <- B2 <- B3 <- 0
  M1 <- M2 <- 0
  
  keepers           <- matrix(0,iters,8)
  colnames(keepers) <- c("Loglike","sigma_e","beta0","beta1","rho","thresh","Acc rate","sigma_a")
  
  # Preprocessing
  
  # list of the sites with non-zero weights for each basis functions
  active <- list()
  for(j in 1:L){
    active[[j]]<-which(B[,j]!=0)
  }
  
  # List of the neighbor set for each knot
  nei  <- list()
  for(j in 1:L){
    nei[[j]] <- which(ADJ[j,]!=0)
  }
  nnei <- rowSums(ADJ)
  siga <- 1/sqrt(nnei)
  
  adj1 <- row(ADJ)[ADJ==1]
  adj2 <- col(ADJ)[ADJ==1]
  low  <- adj1<adj2
  adj1 <- adj1[low]
  adj2 <- adj2[low]
  
  # Eigenvalues of the spatial adjacency matrix
  ev   <- diag(1/sqrt(nnei))%*%as.matrix(ADJ)%*%diag(1/sqrt(nnei)) 
  ev   <- eigen(ev)$values
  
  # Discrete uniform prior for the CAR dependence parameter
  nr       <- 100
  rho.grid <- seq(0.001,0.999,length=nr)
  rho.grid <- qbeta(rho.grid,ar,br)
  eee      <- outer(rho.grid,ev)
  eee      <- 0.5*rowSums(log(1-eee))
  
  weight_grid <- NULL
  for(j in 1:nr){
    weight_grid <- cbind(weight_grid,get.scale(B,ADJ,rho.grid[j]))
  }
  rho.g    <- round(0.75*nr)
  rho      <- rho.grid[rho.g]
  w        <- weight_grid[,rho.g]
  rm(ADJ,low,ev)
  
  # Meen at the initial values
  Z     <- B%*%alpha
  b     <- g_thresh(Z,thresh,w)
  Xb    <- X%*%b
  
  for(iter in 1:iters){
    
    for(ttt in 1:thin){
      
      if(binary){
        MMM <- beta[1]+beta[2]*Xb
        
        #lp  <- pnorm(LOW,MMM,1) 
        #up  <- pnorm(HIGH,MMM,1)  
        #U   <- lp+(up-lp)*runif(n)
        
        #U[U==0] <- 1e-7
        #U[U==1] <- 0.9999999
        
        #Y   <- qnorm(U,MMM,1)
        
        Y  <- rtruncnorm(n,a=LOW,b=HIGH,mean=MMM,sd=1)
      }
      
      # update beta
      
      oldbeta <- beta
      XXX     <- cbind(1,Xb)
      beta    <- update_beta(oldbeta,XXX,Y,taue,eps)
      
      # Update residual variance
      
      SSE  <- sum((Y-beta[1]-beta[2]*Xb)^2) + sum(eps*beta^2)
      taue <- rgamma(1,(n+2)/2+ae,SSE/2+be)
      sige <- 1/sqrt(taue)
      
      # Update basis coefficients (MH with prior as candidate):
      
      old   <- alpha
      
      up_co <- updateCoef(L,active,alpha,nei,rho,siga,Z,B,thresh,w,Xb,X,b,Y,beta,sige,taua)
      alpha <- up_co$u_alpha
      Z     <- up_co$u_Z
      b     <- up_co$u_b
      Xb    <- up_co$u_Xb
      
      # Update the variance of CAR coefficients
      
      R1   <- updateTaua(X,Y,Z,thresh,w,beta,sige,taua,b,Xb)
      taua <- R1$taua
      b    <- R1$b
      Xb   <- R1$Xb
      
      # Update CAR dependence parameter
      
      R2  <- updateRho(X,Y,Z,alpha,adj1,adj2,nr,weight_grid,thresh,beta,sige,taua,rho.grid,eee,
                       rho.g,rho,w,b,Xb)
      rho.g <- R2$rho.g
      rho   <- R2$rho
      w     <- R2$w
      b     <- R2$b
      Xb    <- R2$Xb
      
      # Update threshold
      
      R3     <- updateThresh(t_min,t_max,X,Y,Z,w,beta,sige,taua,thresh,b,Xb)
      thresh <- R3$thresh
      b      <- R3$b
      Xb     <- R3$Xb
      
    }
    
    
    # Keep track of stuff:
    
    if(iter>burn){
      nnn <- iters-burn
      BBB <- beta[2]*b
      B1  <- B1 + BBB/nnn
      B2  <- B2 + BBB*BBB/nnn
      B3  <- B3 + (BBB!=0)/nnn
      
      if(!is.null(Xp)){
        MMM <- beta[1]+beta[2]*Xp%*%b
        if(binary){MMM<-pnorm(MMM)}
        M1  <- M1 + MMM/nnn
        M2  <- M2 + MMM*MMM/nnn
      }
    }
    
    keepers[iter,]  <- c(log.like(Y,beta,Xb,sige),1/sqrt(taue),
                         beta,rho,thresh,mean(old!=alpha),taua)
    
    if(show_iters>0){
      if(iter%%show_iters==0){
        cat("iter ",iter,"\n")
        flush.console()
      }
    }
    
  }        
  
  output<-list(beta.mn=B1,
               beta.var=B2-B1^2,
               beta.nz=B3,
               fit.mn=M1,
               fit.var=M2-M1^2,                 
               keepers=keepers)
  
  return(output)}


############################################################################



#######################################################################
## STGP plots
##
## inputs
## 
## coords := voxels
## img    := data
## dim    := dimension of the vector space
## show   := type of the plots
## split  := location of the polt and split of the screen
## more   := A logical specifying whether more plots will follow on this page.
##
#######################################################################

plot3d.STGP <- function(coords,img,xlab="",ylab="",zlab="",main="",
                        dim=2,show=c("level","cloud"),split=NULL,more=FALSE){
  
  my.cols=c("#000080","#000083","#000087","#00008B","#00008F","#000093","#000097","#00009B",
            "#00009F","#0000A3","#0000A7","#0000AB","#0000AF","#0000B3","#0000B7","#0000BB",
            "#0000BF","#0000C3","#0000C7","#0000CB","#0000CF","#0000D3","#0000D7","#0000DB",
            "#0000DF","#0000E3","#0000E7","#0000EB","#0000EF","#0000F3","#0000F7","#0000FB",
            "#0004FF","#0008FF","#000CFF","#0010FF","#0014FF","#0018FF","#001CFF","#0020FF",
            "#0024FF","#0028FF","#002CFF","#0030FF","#0034FF","#0038FF","#003CFF","#0040FF",
            "#0044FF","#0048FF","#004CFF","#0050FF","#0054FF","#0058FF","#005CFF","#0060FF",
            "#0064FF","#0068FF","#006CFF","#0070FF","#0074FF","#0078FF","#007CFF","#0080FF",
            "#0083FF","#0087FF","#008BFF","#008FFF","#0093FF","#0097FF","#009BFF","#009FFF",
            "#00A3FF","#00A7FF","#00ABFF","#00AFFF","#00B3FF","#00B7FF","#00BBFF","#00BFFF",
            "#00C3FF","#00C7FF","#00CBFF","#00CFFF","#00D3FF","#00D7FF","#00DBFF","#00DFFF",
            "#00E3FF","#00E7FF","#00EBFF","#00EFFF","#00F3FF","#00F7FF","#00FBFF","#00FFFF",
            "#04FFFB","#08FFF7","#0CFFF3","#10FFEF","#14FFEB","#18FFE7","#1CFFE3","#20FFDF",
            "#24FFDB","#28FFD7","#2CFFD3","#30FFCF","#34FFCB","#38FFC7","#3CFFC3","#40FFBF",
            "#44FFBB","#48FFB7","#4CFFB3","#50FFAF","#54FFAB","#58FFA7","#5CFFA3","#60FF9F",
            "#64FF9B","#68FF97","#6CFF93","#70FF8F","#74FF8B","#78FF87","#7CFF83","#80FF80",
            "#83FF7C","#87FF78","#8BFF74","#8FFF70","#93FF6C","#97FF68","#9BFF64","#9FFF60",
            "#A3FF5C","#A7FF58","#ABFF54","#AFFF50","#B3FF4C","#B7FF48","#BBFF44","#BFFF40",
            "#C3FF3C","#C7FF38","#CBFF34","#CFFF30","#D3FF2C","#D7FF28","#DBFF24","#DFFF20",
            "#E3FF1C","#E7FF18","#EBFF14","#EFFF10","#F3FF0C","#F7FF08","#FBFF04","#FFFF00",
            "#FFFB00","#FFF700","#FFF300","#FFEF00","#FFEB00","#FFE700","#FFE300","#FFDF00",
            "#FFDB00","#FFD700","#FFD300","#FFCF00","#FFCB00","#FFC700","#FFC300","#FFBF00",
            "#FFBB00","#FFB700","#FFB300","#FFAF00","#FFAB00","#FFA700","#FFA300","#FF9F00",
            "#FF9B00","#FF9700","#FF9300","#FF8F00","#FF8B00","#FF8700","#FF8300","#FF8000",
            "#FF7C00","#FF7800","#FF7400","#FF7000","#FF6C00","#FF6800","#FF6400","#FF6000",
            "#FF5C00","#FF5800","#FF5400","#FF5000","#FF4C00","#FF4800","#FF4400","#FF4000",
            "#FF3C00","#FF3800","#FF3400","#FF3000","#FF2C00","#FF2800","#FF2400","#FF2000",
            "#FF1C00","#FF1800","#FF1400","#FF1000","#FF0C00","#FF0800","#FF0400","#FF0000",
            "#FB0000","#F70000","#F30000","#EF0000","#EB0000","#E70000","#E30000","#DF0000",
            "#DB0000","#D70000","#D30000","#CF0000","#CB0000","#C70000","#C30000","#BF0000",
            "#BB0000","#B70000","#B30000","#AF0000","#AB0000","#A70000","#A30000","#9F0000",
            "#9B0000","#970000","#930000","#8F0000","#8B0000","#870000","#830000","#800000")
  
  ## 2D
  if (dim==2){
    
    ## cloud
    
    if (show[1]=="cloud"){
      fig = cloud(img~coords$x*coords$y,xlab=xlab,ylab=ylab,zlab=list(zlab,rot=90),main=main,
                  scales=list(z=list(arrows=FALSE)))}
    
    ## levelplot
    
    else{
      fig = levelplot(img~coords$x*coords$y, col.regions=my.cols, aspect=1,
                      xlab=xlab, ylab=ylab, main=main, 
                      at=seq(min(img),max(img),length.out =length(my.cols)-1))}
    
  }
  
  ## 3D
  
  if (dim==3){
    
    ## cloud
    
    if (show[1]=="cloud"){
      Coord_z = shingle(coords$z, intervals=quantile(coords$z, c(0,1/6,2/6,3/6,4/6,5/6,1),type=1))
      fig = cloud(img~coords$x*coords$y|Coord_z,xlab=xlab,ylab=ylab,zlab=list(zlab,rot=90),
                  main=main,layout=c(7,1),scales=list(z=list(arrows=FALSE)))}
    
    ## levelplot
    
    else{
      Coord_z = shingle(coords$z, intervals=quantile(coords$z, c(0,1/6,2/6,3/6,4/6,5/6,1),type=1))
      #Coord_z = equal.count(coords$z, 5, overlap=0)
      #Coord_z = cut(coords$z, 5)
      fig = levelplot(img~coords$x*coords$y|Coord_z,col.regions=my.cols, aspect=1,
                      xlab=xlab, ylab=ylab, main=main, layout=c(7,1), 
                      at=seq(min(img),max(img),length.out =length(my.cols)-1))}
    
  }
  
  plot(fig,more=more,split=split)
} 

############################################################################



#######################################################################
## STGP training analysis
##
## inputs
## 
## beta   := true beta
## fit    := results of fit
## thresh := thresh of judgement
##
#######################################################################

perform.STGP <- function(beta,fit,thresh=0.5){
  
  p = length(beta)
  
  ## MSE
  
  beta.mse <- sum((fit$beta.var)+(fit$beta.mn-beta)^2) / p
  beta.mse <- list(beta.mse = beta.mse)
  
  ## sensitivity v.s. spacifity
  
  beta.train = ifelse(fit$beta.nz < thresh, 0, 1)
  
  True.Positive = sum((beta==1)&(beta==beta.train))
  False.Positive = sum((beta==0)&(beta!=beta.train))
  True.Negative = sum((beta==0)&(beta==beta.train))
  False.Negative = sum((beta==1)&(beta!=beta.train))
  
  train.table = matrix(c(True.Positive,False.Negative,False.Positive,True.Negative),ncol=2)
  colnames(train.table) = c("True.Pos","True.Neg")
  rownames(train.table) = c("Pred.Pos","Pred.Neg") 
  train.table = as.table(train.table)
  
  sensitivity = True.Positive / (True.Positive+False.Negative)
  specificity = True.Negative / (True.Negative+False.Positive)
  PPV          = True.Positive / (True.Positive+False.Positive)
  NPV         = True.Negative / (True.Negative+False.Negative)
  
  train.value = c(sensitivity,specificity,PPV,NPV)
  names(train.value) = c("sensitivity","specificity","PPV","NPV")
  train.value = as.table(train.value)
  
  results = list(beta.mse=beta.mse,
                 table=train.table,
                 Sens=train.value)
}

